/* Copyright (C) 1999-2010 Thomas S. Glascock */
/* www.soomka.com */

#include <pthread.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <math.h>
#include <sys/time.h>
#include <time.h>

#define MINUTE 1
#define HOUR 2
#define SECOND 3
#define VERSION "0.2.3F"

#define ALARMLENGTH (45)
#define SNOOZELENGTH (5)

char key;
int getkeys;

void *keygetcher(void *)
{
	extern char key;

	while (1)
	{
		pthread_testcancel();
		key=getch();
	}
	return 0;
}

inline int is_even(int numma)
{
	return (!(numma & 1));
}

inline float htod(int hours, int minutes)
{
	return	 	(
			((float)(hours>12?hours-12:hours) +
			((float)minutes / 60.0) // +
			) * 30.0 );
}

inline float mtod(int minutes, int seconds)
{
	return	 	(
			((float)minutes + 
			((float)seconds / 60.0) // +
			) * 6.0 );
}

inline float stod(int seconds, int microseconds)
{
	return	 	(
			((float)seconds +
			((float)microseconds  / 1000000.0)
			) * 6.0 );
}

class timex
{
	public:
		timex();
		~timex();
		void PrintFace();
		void PrintHands(float,float,float);
		void PrintHour(int,int);
		void PrintMinute(int,int);
		void PrintSecond(int,int);
		void RunClock();
		void Info();
		void Slide();
		void RedoWindows();
		void SaveSettings();
		void LoadSettings();

	private:
		int snooze;
		int silenced;
		int tripped;
		int alarm_on;
		int alarm_hour;
		int alarm_minute;
		int alarm_ringing;
		int alarm_length;
		int snooze_length;
		int tmptimeout;
		int slider;
		int color;
		int main_lines;
		int main_cols;
		int slide_lines;
		int slide_cols;
		int slide_x;
		int slide_y;
		int slide_x_direction;
		int slide_y_direction;
		int slide_lock;
		long alarmend;
		long snoozeend;
		long tmpend;
		float pct[5];
		float pi;
		float aspect;
		float pi_six;
		float pi_180;
		float y_half;
		float x_half;
		float ymult;
		float prevhour;
		float prevminute;
		float prevsecond;
		char *junk;
		char *symbol[5];
		WINDOW *face_win;
		WINDOW *hand_win;
		WINDOW *hour_win;
		WINDOW *minute_win;
		WINDOW *second_win;
};

timex::timex()
{
	pi=3.141592654;
	aspect=1.333333333;
	junk=(char *)calloc(1,256);

	symbol[1]=(char *)calloc(1,1);
	symbol[2]=(char *)calloc(1,1);
	symbol[3]=(char *)calloc(1,1);

/*
	sprintf(symbol[1],"m");
	sprintf(symbol[2],"h");
	sprintf(symbol[3],"s");
*/

	*symbol[1]='m';
	*symbol[2]='h';
	*symbol[3]='s';

	pct[HOUR]=60.0/100.0;
	pct[MINUTE]=95.0/100.0;
	pct[SECOND]=95.0/100.0;

	alarm_on=0;
	alarm_hour=0;
	alarm_minute=0;
	alarm_ringing=0;
	snooze=0;
	silenced=0;
	tripped=0;
	alarmend=0;
	snoozeend=0;
	tmpend=0;
	tmptimeout=0;
	slider=0;
        color=0;
	slide_x_direction=1;
	slide_y_direction=1;
	slide_lock=0;

	alarm_length=ALARMLENGTH;
	snooze_length=SNOOZELENGTH;

	initscr();
	start_color();
        assume_default_colors(-1, -1);
	init_pair(1, COLOR_YELLOW, -1); // second hand
	init_pair(2, COLOR_GREEN, -1); // minute hand
	init_pair(3, COLOR_BLUE, -1); // hour hand
	init_pair(4, COLOR_WHITE, -1); // face
	attrset(A_NORMAL);
	curs_set(0);
	halfdelay(0);
	noecho();
	main_lines=LINES;
	main_cols=COLS;
	hand_win=newwin(main_lines,main_cols,0,0);
	face_win=newwin(main_lines,main_cols,0,0); 
	hour_win=newwin(main_lines,main_cols,0,0);
	minute_win=newwin(main_lines,main_cols,0,0);
	second_win=newwin(main_lines,main_cols,0,0);
	prevhour=-1;
	prevminute=-1;
	prevsecond=-1;
}

timex::~timex()
{
	free(junk);
	curs_set(1);
	nocbreak();
	echo();
	nl();
	endwin();
}

void timex::SaveSettings()
{
	FILE *clockywockfile;
	char *filename;
	filename=(char *)calloc(1,256);
	strcat(filename,getenv("HOME"));
	strcat(filename,"/.clockywock");
	clockywockfile=fopen(filename,"w");
	fprintf(clockywockfile,"%d\n",alarm_on);
	fprintf(clockywockfile,"%d\n",alarm_hour);
	fprintf(clockywockfile,"%d\n",alarm_minute);
	fprintf(clockywockfile,"%d\n",alarm_ringing);
	fprintf(clockywockfile,"%d\n",alarm_length);
	fprintf(clockywockfile,"%d\n",snooze_length);
	fprintf(clockywockfile,"%d\n",slider);
	fprintf(clockywockfile,"%d\n",color);
	fprintf(clockywockfile,"%c\n",*symbol[1]);
	fprintf(clockywockfile,"%c\n",*symbol[2]);
	fprintf(clockywockfile,"%c\n",*symbol[3]);
	fclose(clockywockfile);
	free(filename);
}

void timex::LoadSettings()
{
	FILE *clockywockfile;
	char *filename;
	filename=(char *)calloc(1,256);
	strcat(filename,getenv("HOME"));
	strcat(filename,"/.clockywock");
	clockywockfile=fopen(filename,"r");
	if (clockywockfile!=NULL)
	{
		fscanf(clockywockfile,(const char *)"%d\n",&alarm_on);
		fscanf(clockywockfile,(const char *)"%d\n",&alarm_hour);
		fscanf(clockywockfile,(const char *)"%d\n",&alarm_minute);
		fscanf(clockywockfile,(const char *)"%d\n",&alarm_ringing);
		fscanf(clockywockfile,(const char *)"%d\n",&alarm_length);
		fscanf(clockywockfile,(const char *)"%d\n",&snooze_length);
		fscanf(clockywockfile,(const char *)"%d\n",&slider);
		fscanf(clockywockfile,(const char *)"%d\n",&color);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[1]);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[2]);
		fscanf(clockywockfile,(const char *)"%c\n",symbol[3]);
		fclose(clockywockfile);
	}
	free(filename);
}

void timex::Slide()
{
	if (slide_y_direction==1)
	{
		if (slide_y<slide_lines)
			slide_y=slide_y+1;
		else
		{
			slide_y_direction=-1;
		}
	}
	else
	{
		if (slide_y>0)
			slide_y=slide_y-1;
		else
		{
			slide_y_direction=1;
		}
	}

	if (slide_x_direction==1)
	{
		if (slide_x<slide_cols)
			slide_x=slide_x+1;
		else
		{
			slide_x_direction=-1;
		}
	}
	else
	{
		if (slide_x>0)
			slide_x=slide_x-1;
		else
		{
			slide_x_direction=1;
		}
	}
	wclear(hand_win);
	wrefresh(hand_win);
	mvwin(face_win,slide_y,slide_x);
	mvwin(hand_win,slide_y,slide_x);
	mvwin(hour_win,slide_y,slide_x);
	mvwin(minute_win,slide_y,slide_x);
	mvwin(second_win,slide_y,slide_x);
}

void timex::Info()
{
	WINDOW *infowin;
	int y,x;
	int infolines,infocols;
	infolines=22;
	infocols=44;

	infowin=newwin(infolines,infocols,int(y_half)-(infolines/2),int(x_half)-(infocols/2));
	getmaxyx(infowin,y,x);
	box(infowin,0,0);

	sprintf(junk,"key       function                value");
	mvwaddstr(infowin,1,2,junk);
	sprintf(junk,"(1,!)     minute hand character:  \"%c\"",*symbol[1]);
	mvwaddstr(infowin,3,2,junk);
	sprintf(junk,"(2,@)     hour hand character:    \"%c\"",*symbol[2]);
	mvwaddstr(infowin,4,2,junk);
	sprintf(junk,"(3,#)     second hand character:  \"%c\"",*symbol[3]);
	mvwaddstr(infowin,5,2,junk);
	sprintf(junk,"(A)       alarm:                   %s",alarm_on==1?"on":"off");
	mvwaddstr(infowin,7,2,junk);
	sprintf(junk,"(h,H;m,M) alarm time:              %2.2d:%2.2d",alarm_hour,alarm_minute);
	mvwaddstr(infowin,9,2,junk);
	sprintf(junk,"(t,T)     alarm length:            %2.2dm",alarm_length);
	mvwaddstr(infowin,11,2,junk);
	sprintf(junk,"(z,Z)     snooze length:           %2.2dm",snooze_length);
	mvwaddstr(infowin,13,2,junk);
	sprintf(junk,"(s,S)     slider mode:             %s",slider==1?"on":"off");
	mvwaddstr(infowin,15,2,junk);
	sprintf(junk,"(c,C)     color mode:              %s",color==1?"on":"off");
	mvwaddstr(infowin,17,2,junk);
	sprintf(junk,"<space>   snoozzze                 %s",snooze==1?"on":"off");
	mvwaddstr(infowin,19,2,junk);
	sprintf(junk,"<enter>   kill alarm");
	mvwaddstr(infowin,20,2,junk);
	wrefresh(infowin);
	overwrite(infowin,hand_win);
	delwin(infowin);
}

void timex::RedoWindows()
{
	wclear(hand_win);
	wclear(face_win);
	wclear(hour_win);
	wclear(minute_win);
	wclear(second_win);

	wrefresh(hand_win);
	wrefresh(face_win);
	wrefresh(hour_win);
	wrefresh(minute_win);
	wrefresh(second_win);

	delwin(hand_win);
	delwin(face_win);
	delwin(hour_win);
	delwin(minute_win);
	delwin(second_win);

	if (slider==1)
	{
		main_lines=(LINES-(LINES/8));
		main_cols=(COLS-(COLS/8));
		slide_lines=LINES-main_lines;
		slide_cols=COLS-main_cols;
		slide_x=0;
		slide_y=0;
	} else {
		main_lines=LINES;
		main_cols=COLS;
	}

	x_half=(main_cols/2.0)+(is_even(main_cols)?-0.5:0.0);
	y_half=(main_lines/2.0)+(is_even(main_lines)?-0.5:0.0);
	pi_six=pi/6.0;
	ymult=(y_half/x_half);
	pi_180=pi/180.0;

	hand_win=newwin(main_lines,main_cols,0,0);
	face_win=newwin(main_lines,main_cols,0,0); 
	hour_win=newwin(main_lines,main_cols,0,0);
	minute_win=newwin(main_lines,main_cols,0,0);
	second_win=newwin(main_lines,main_cols,0,0);

	if (color) {
                assume_default_colors(COLOR_WHITE,COLOR_BLACK);
//		wattron(face_win,COLOR_PAIR(4));
		wattron(hour_win,COLOR_PAIR(3));
		wattron(minute_win,COLOR_PAIR(2));
		wattron(second_win,COLOR_PAIR(1));

		wattron(hour_win,A_BOLD);
		wattron(minute_win,A_BOLD);
		wattron(second_win,A_BOLD);
	} else {
                assume_default_colors(-1,-1);
//		wattroff(face_win,COLOR_PAIR(4));
		wattroff(hour_win,COLOR_PAIR(3));
		wattroff(minute_win,COLOR_PAIR(2));
		wattroff(second_win,COLOR_PAIR(1));

		wattroff(hour_win,A_BOLD);
		wattroff(minute_win,A_BOLD);
		wattroff(second_win,A_BOLD);
	}

	wrefresh(hand_win);
	wrefresh(face_win);
	wrefresh(hour_win);
	wrefresh(minute_win);
	wrefresh(second_win);

	prevhour=-1;
	prevminute=-1;
	prevsecond=-1;
}

void timex::PrintFace()
{
	int m;

	wclear(face_win);
	wattron(face_win,A_STANDOUT);

	for (m=0;m<=11;++m)
	{
		mvwaddstr(face_win,
			(int)(y_half + (  y_half * (-cos(pi_six*m)))),
			(int)(x_half + (( x_half * ( sin(pi_six*m)))) / aspect),
			" ");
	}
	wattroff(face_win,A_STANDOUT);
	sprintf(junk,"clockywock v.%s", VERSION);
	mvwaddstr(face_win,main_lines-1,0,junk);
	mvwaddstr(face_win,(int)(y_half),(int)(x_half),"O");
}

void timex::PrintHour(int hour, int minute)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	werase(hour_win);

	unit=htod(hour,minute)*pi_180;

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(x_half * pct[HOUR]);

	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(hour_win, (int)(y_half+(x*-cosz)), (int)(x_half+(x*sinz)), symbol[HOUR]);
	}
	prevminute=minute;
}

void timex::PrintMinute(int minute, int second)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	werase(minute_win);

	unit=mtod(minute,second)*pi_180;

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(x_half * pct[MINUTE]);
	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(minute_win, (int)(y_half+(x*-cosz)), (int)(x_half+(x*sinz)), symbol[MINUTE]);
	}
	prevsecond=second;
}

void timex::PrintSecond(int second, int microsecond)
{
	float x;
	float sinz;
	float cosz;
	float handlength;
	float unit;

	unit=stod(second,microsecond)*pi_180;

	werase(second_win);

	cosz=((cos(unit))*ymult);
	sinz=((sin(unit))/aspect);;
	handlength=(int)(x_half * pct[SECOND]);
	for (x=1;x!=handlength;++x)
	{
		mvwaddstr(second_win, (int)(y_half+(x*-cosz)), (int)(x_half+(x*sinz)), symbol[SECOND]);
	}
	prevsecond=second;
}


// void timex::PrintHands(float hourdeg, float minutedeg, float seconddeg)
// {
// 	float x;
// 	float sinz;
// 	float cosz;
// 	float handlength;
// 	float unit[4];
// 
// 	int increment;
// 
// 	overwrite(face_win,hand_win);
// 
// 	unit[HOUR]=hourdeg*pi_180;
// 	unit[MINUTE]=minutedeg*pi_180;
// 	unit[SECOND]=seconddeg*pi_180;
// 
// 	attrset(A_NORMAL);
// 	
// 	for (increment=MINUTE;increment<=SECOND;++increment)
// 	{
// 		cosz=((cos(unit[increment]))*ymult);
// 		sinz=((sin(unit[increment]))/aspect);;
// 		handlength=(int)(x_half * pct[increment]);
// 		for (x=1;x!=handlength;++x)
// 		{
// 			mvwaddstr(hand_win, (int)(y_half+(x*-cosz)), (int)(x_half+(x*sinz)),
// 			symbol[increment]);
// 		}
// 	}
// 	overlay(face_win,hand_win);
// 
// 	if (tmptimeout==1)
// 	{
// 		Info();
// 	}
// }

void timex::RunClock()
{
	extern char key;
	extern int getkeys;

	int prevlines=0;
	int prevcols=0;

	struct timeval td;
	struct tm *ct;

	pthread_t keythread;

	ct=(struct tm*)NULL;

	LoadSettings();

	if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);

	RedoWindows();

	for (;;)
	{
		if ((COLS!=prevcols)||(LINES!=prevlines))
		{
			pthread_cancel(keythread);
			RedoWindows();
			if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);
			PrintFace();
		}

		prevcols=COLS;
		prevlines=LINES;

		gettimeofday(&td,(struct timezone *)NULL);
		ct=localtime(&td.tv_sec);

		overwrite(face_win,hand_win);

		/* if (ct->tm_min!=prevminute) PrintHour(ct->tm_hour,ct->tm_min); */
		PrintHour(ct->tm_hour,ct->tm_min);

		overlay(hour_win,hand_win);

		/* if (( ((ct->tm_sec%15)==0) && (ct->tm_sec!=prevsecond) ) || (prevsecond==-1)) PrintMinute(ct->tm_min,ct->tm_sec); */

                PrintMinute(ct->tm_min,ct->tm_sec);

		overlay(minute_win,hand_win);

		PrintSecond(ct->tm_sec,td.tv_usec);

		overlay(second_win,hand_win);

		overlay(face_win,hand_win);

		if (tmptimeout==1)
		{
			Info();
		}

		if (alarm_on==1)
		{
			if (silenced==0)
			{
				if (alarmend > td.tv_sec)
				{
					if (snoozeend < td.tv_sec)
					{
						alarm_ringing=1;
						snooze=0;
					} else
						alarm_ringing=0;
				} else
					alarmend=snooze=alarm_ringing=tripped=0;
			} else
			{
				if (alarmend < td.tv_sec)
					silenced=tripped=alarm_ringing=0;
				else
				{
					alarm_ringing=0;
				}
			}
			if (((ct->tm_hour==alarm_hour)&&(ct->tm_min==alarm_minute)))
			{
				if (tripped==0)
				{
					alarmend=td.tv_sec+(alarm_length * 60) ;
					tripped=1;
				}
			}
			if (alarm_ringing==1) { beep(); flash(); }
		}

		if (tmpend<td.tv_sec) { tmpend=tmptimeout=0; }

		// key is fetched in a separate thread

		if ((key!=-1))
		{
			tmpend=td.tv_sec+5;
			tmptimeout=1;

			switch (key)
			{
				case 'a':
				case 'A':
					alarm_on=(1-(alarm_on)); break;
	
				case 'h':
					alarm_hour = (alarm_hour)<23 ? ++alarm_hour : 0; break;
				case 'H':
					alarm_hour = (alarm_hour)>0 ? --alarm_hour : 23; break;
	
				case 'm':
					alarm_minute = (alarm_minute)<59 ? ++alarm_minute : 0; break;
				case 'M':
					alarm_minute = (alarm_minute)>0 ? --alarm_minute : 59; break;
	
				case 't':
					alarm_length = (alarm_length)<60 ? ++alarm_length : 15; break;
				case 'T':
					alarm_length = (alarm_length)>15 ? --alarm_length : 60; break;
	
				case 'z':
					snooze_length = (snooze_length)<30 ? ++snooze_length : 3; break;
				case 'Z':
					snooze_length = (snooze_length)>3 ? --snooze_length : 30; break;
	
				case '1':
					*symbol[1] = (*symbol[1])<126 ? ++*symbol[1] : 33; prevsecond=-1; break;
	
				case '!':
					*symbol[1] = (*symbol[1])>33 ? --*symbol[1] : 126; prevsecond=-1; break;
	
				case '2':
					*symbol[2] = (*symbol[2])<126 ? ++*symbol[2] : 33; prevminute=-1; break;
	
				case '@':
					*symbol[2] = (*symbol[2])>33 ? --*symbol[2] : 126; prevminute=-1; break;
	
				case '3':
					*symbol[3] = (*symbol[3])<126 ? ++*symbol[3] : 33; break;
	
				case '#':
					*symbol[3] = (*symbol[3])>33 ? --*symbol[3] : 126; break;

				case '/':
					tmpend=td.tv_sec; break;
	
				case 'q':
				case 'Q':
				{
					clear();
					wrefresh(hand_win);
					SaveSettings();
					return;
					break;
				}
	
				case 'c':
				case 'C':
					color=1-color;

					if (color) {
                				assume_default_colors(COLOR_WHITE,COLOR_BLACK);
//						wattron(face_win,COLOR_PAIR(4));
						wattron(hour_win,COLOR_PAIR(3));
						wattron(minute_win,COLOR_PAIR(2));
						wattron(second_win,COLOR_PAIR(1));

						wattron(hour_win,A_BOLD);
						wattron(minute_win,A_BOLD);
						wattron(second_win,A_BOLD);
					} else {
                				assume_default_colors(-1,-1);
//						wattroff(face_win,COLOR_PAIR(4));
						wattroff(hour_win,COLOR_PAIR(3));
						wattroff(minute_win,COLOR_PAIR(2));
						wattroff(second_win,COLOR_PAIR(1));

						wattroff(hour_win,A_BOLD);
						wattroff(minute_win,A_BOLD);
						wattroff(second_win,A_BOLD);
					}
					wclear(hand_win);
					wrefresh(hand_win);
					break;
				case 's':
				case 'S':
					wclear(hand_win);
					wrefresh(hand_win);
					slider=1-slider;
					pthread_cancel(keythread);
					RedoWindows();
					if (getkeys == 1) pthread_create(&keythread,NULL,keygetcher,NULL);
					PrintFace(); break;
	
				case ' ': 
					if ((alarm_on==1) && (alarm_ringing==1))
					{
						snooze=1;
						snoozeend=td.tv_sec+ (snooze_length * 60);
					}
					break;
	
				case 10:
					if (alarm_on==1) silenced=1; break;
	
				case 13:
					if (alarm_on==1) silenced=1; break;
	
			}
			key=-1;
		}
		if ((slider==1)  && (slide_lock==0) && (ct->tm_sec==0))
		{
			Slide();
			slide_lock=1;
		} 

		if ((ct->tm_sec%5)==1) slide_lock=0;

		if (slider==1) wrefresh(hand_win);

		wrefresh(hand_win);

		usleep(10000);
	}
}

int main(int argc, char **argv)
{
	extern char key;
	/* extern char* optarg; */
	/* extern int optind, opterr, optopt; */
	int q;
	extern int getkeys;

	getkeys=1;

	key=-1;

	timex slimex;

	while (1) {
		q = getopt (argc, argv, "hk");
		printf ("q is [%d]\n",q);

		if (q == -1)
			break;

		switch (q) {
			case 'h':
				printf ("help\n");
				exit(0);
				break;
			case 'k':
				getkeys = 0;
				break;
		}
	}

	slimex.RunClock();
}
